/****
DIAMOND protein aligner
Copyright (C) 2013-2018 Benjamin Buchfink <buchfink@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
****/

#pragma once
#include "stream_entity.h"

struct FileSource : public StreamEntity
{
	FileSource(const std::string &file_name);
	FileSource(const std::string &file_name, FILE *file);
	virtual void rewind() override;
	virtual void seek(int64_t p, int origin) override;
	virtual int64_t tell() override;
	virtual size_t read(char *ptr, size_t count) override;
	virtual void close() override;
	virtual int64_t file_size() override;
	virtual bool eof() override;
	virtual const std::string& file_name() const override
	{
		return file_name_;
	}
	virtual FILE* file() override
	{
		return f_;
	}
	//void putback(char c);
	~FileSource()
	{}
protected:
	FILE *f_;
	const std::string file_name_;
};